/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::NavigationMemory
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Visu.h"
#include <Eigen/src/Geometry/Translation.h>

#include <SimoxUtility/color/Color.h>
#include <SimoxUtility/color/cmaps/colormaps.h>

#include "RobotAPI/components/ArViz/Client/Elements.h"
#include "RobotAPI/components/ArViz/Client/Layer.h"
#include "RobotAPI/components/ArViz/Client/elements/Color.h"
#include "RobotAPI/components/ArViz/Client/elements/Robot.h"
#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>

#include "armarx/navigation/conversions/eigen.h"
#include "armarx/navigation/human/aron/Human.aron.generated.h"
#include "armarx/navigation/human/aron_conversions.h"
#include "armarx/navigation/human/types.h"
#include <armarx/navigation/algorithms/aron/Costmap.aron.generated.h>
#include <armarx/navigation/algorithms/aron_conversions.h>
#include <armarx/navigation/core/Graph.h>
#include <armarx/navigation/core/aron/Graph.aron.generated.h>
#include <armarx/navigation/core/aron/Location.aron.generated.h>
#include <armarx/navigation/graph/Visu.h>
#include <armarx/navigation/human/aron/Human.aron.generated.h>


namespace armarx::navigation::memory
{

    Visu::Visu(viz::Client arviz,
               const armem::server::wm::CoreSegment& locSegment,
               const armem::server::wm::CoreSegment& graphSegment,
               const armem::server::wm::CoreSegment& costmapSegment,
               const armem::server::wm::CoreSegment& humanSegment) :
        arviz(arviz),
        locSegment(locSegment),
        graphSegment(graphSegment),
        costmapSegment(costmapSegment),
        humanSegment(humanSegment),
        visu(std::make_unique<graph::GraphVisu>())
    {
    }


    Visu::~Visu()
    {
    }


    void
    Visu::drawLocations(std::vector<viz::Layer>& layers, bool enabled)
    {
        using namespace armem::server;

        viz::Layer& layer = layers.emplace_back(arviz.layer(locSegment.id().str()));
        if (enabled)
        {
            std::map<armem::MemoryID, location::arondto::Location> locations;
            locSegment.doLocked(
                [&]()
                {
                    locSegment.forEachEntity(
                        [&](const wm::Entity& entity)
                        {
                            if (const wm::EntityInstance* instance = entity.findLatestInstance())
                            {
                                locations[entity.id()].fromAron(instance->data());
                            }
                        });
                });

            for (auto& [id, location] : locations)
            {
                visu->vertex->draw(layer, id.str(), core::Pose(location.globalRobotPose));
            }
        }
    }


    void
    Visu::drawGraphs(std::vector<viz::Layer>& layers, bool enabled)
    {
        using namespace armem::server;

        std::map<armem::MemoryID, core::Graph> graphs;
        graphSegment.doLocked(
            [&]()
            {
                graphSegment.forEachEntity(
                    [&](const wm::Entity& entity)
                    {
                        core::Graph& graph = graphs[entity.id()];
                        if (enabled)
                        {
                            if (const wm::EntityInstance* instance = entity.findLatestInstance())
                            {
                                navigation::core::arondto::Graph aron;
                                aron.fromAron(instance->data());
                                fromAron(aron, graph);
                            }
                        }
                        // else: empty layer
                    });
            });

        for (auto& [id, graph] : graphs)
        {
            viz::Layer& layer = layers.emplace_back(arviz.layer(id.str()));
            if (enabled)
            {
                core::resolveLocations(graph, locSegment);
                visu->draw(layer, graph);
            }
            // else: clear layer
        }
    }

    namespace
    {

        void
        visualize(const algorithms::Costmap& costmap, viz::Layer& layer, const std::string& name)
        {
            const auto cmap = simox::color::cmaps::viridis();
            const float vmax = costmap.getGrid().array().maxCoeff();

            const auto asColor = [&cmap, &vmax](const float distance) -> viz::data::Color
            {
                const auto color = cmap.at(distance, 0.F, vmax);
                return {color.a, color.r, color.g, color.b};
            };

            const std::int64_t cols = costmap.getGrid().cols();
            const std::int64_t rows = costmap.getGrid().rows();

            auto mesh = viz::Mesh(name);

            std::vector<std::vector<Eigen::Vector3f>> vertices;
            std::vector<std::vector<viz::data::Color>> colors;

            for (int r = 0; r < rows; r++)
            {
                auto& verticesRow = vertices.emplace_back(cols);
                auto& colorsRow = colors.emplace_back(cols);

                for (int c = 0; c < cols; c++)
                {
                    verticesRow.at(c) = conv::to3D(costmap.toPositionGlobal({r, c}));
                    colorsRow.at(c) = asColor(costmap.getGrid()(r, c));
                }
            }

            mesh.grid2D(vertices, colors);

            layer.add(mesh);
        }

        void
        visualize(const human::Humans& humans, viz::Layer& layer)
        {

            const Eigen::Translation3f human_T_mmm(Eigen::Vector3f{0,0, 1000});

            ARMARX_INFO << "Visualizing " << humans.size() << " humans";
            for (const auto& human : humans)
            {
                // viz::Cylinder cylinder(std::to_string(layer.size()));
                // cylinder.fromTo(conv::to3D(human.pose.translation()),
                //                 conv::to3D(human.pose.translation()) + Eigen::Vector3f{0, 0, 10});


                // cylinder.color(simox::Color::orange());
                // cylinder.radius(300);
                // layer.add(cylinder);


                viz::Robot mmm(std::to_string(layer.size()));
                mmm.file("RobotAPI", "RobotAPI/robots/MMM/mmm.xml");
                mmm.pose(conv::to3D(human.pose) * human_T_mmm);
                mmm.scale(1.7); // 1.7m 
                mmm.overrideColor(viz::Color::orange(255, 100));
                layer.add(mmm);

            }
        }

    } // namespace

    void
    Visu::drawCostmaps(std::vector<viz::Layer>& layers, bool enabled)
    {
        if (not enabled)
        {
            return;
        }

        std::map<std::string, std::vector<std::pair<std::string, algorithms::Costmap>>>
            namedProviderCostmaps;

        costmapSegment.doLocked(
            [&]()
            {
                using namespace armem::server;

                costmapSegment.forEachEntity(
                    [&](const wm::Entity& entity)
                    {
                        if (const wm::EntityInstance* instance = entity.findLatestInstance())
                        {
                            navigation::algorithms::Costmap costmap =
                                algorithms::fromAron(*instance);

                            namedProviderCostmaps[instance->id().providerSegmentName].emplace_back(
                                instance->id().entityName, std::move(costmap));
                        }
                    });
            });

        for (const auto& [providerName, namedCostmaps] : namedProviderCostmaps)
        {
            viz::Layer& layer = layers.emplace_back(arviz.layer("costmaps_" + providerName));
            for (const auto& [name, costmap] : namedCostmaps)
            {
                visualize(costmap, layer, name);
            }
        }
    }

    void
    Visu::drawHumans(std::vector<viz::Layer>& layers, bool enabled)
    {
        if (not enabled)
        {
            return;
        }

        std::map<std::string, navigation::human::Humans> namedProviderHumans;

        humanSegment.doLocked(
            [&]()
            {
                using namespace armem::server;

                humanSegment.forEachEntity(
                    [&](const wm::Entity& entity)
                    {
                        entity.getLatestSnapshot().forEachInstance(
                            [&namedProviderHumans](const armarx::armem::wm::EntityInstance& instance)
                            {
                                const auto dto =
                                    navigation::human::arondto::Human::FromAron(instance.data());

                                navigation::human::Human human;
                                fromAron(dto, human);

                                namedProviderHumans[instance.id().providerSegmentName]
                                    .emplace_back(std::move(human));
                            });
                    });
            });

        for (const auto& [providerName, humans] : namedProviderHumans)
        {
            viz::Layer& layer = layers.emplace_back(arviz.layer("humans_" + providerName));
            visualize(humans, layer);
        }
    }


} // namespace armarx::navigation::memory
