#include "TebObstacleManager.h"

#include <ArmarXCore/core/logging/Logging.h>

#include <armarx/navigation/conversions/eigen.h>
#include <armarx/navigation/local_planning/ros_conversions.h>
#include <teb_local_planner/extension/obstacles/EllipseObstacle.h>
#include <teb_local_planner/obstacles.h>

namespace armarx::navigation::local_planning
{

    void
    TebObstacleManager::clear()
    {
        container.clear();
        visualizationIndex = 0;
    }

    size_t
    TebObstacleManager::size()
    {
        return container.size();
    }

    void
    TebObstacleManager::addBoxObstacle(const VirtualRobot::BoundingBox& bbox, viz::Layer* visLayer)
    {
        auto obst = boost::make_shared<teb_local_planner::PolygonObstacle>();

        const Eigen::Vector2d min = conv::toRos(bbox.getMin());
        const Eigen::Vector2d max = conv::toRos(bbox.getMax());

        obst->pushBackVertex(min);
        obst->pushBackVertex(min.x(), max.y());
        obst->pushBackVertex(max);
        obst->pushBackVertex(max.x(), min.y());

        obst->finalizePolygon();
        container.push_back(obst);

        // visualize bounding box if layer is available
        if (visLayer != nullptr)
        {
            const Eigen::Vector3f min3d = conv::fromRos(min);
            const Eigen::Vector3f max3d = conv::fromRos(max);

            visLayer->add(viz::Polygon("polygon_" + std::to_string(visualizationIndex++))
                              .addPoint(min3d)
                              .addPoint(Eigen::Vector3f(min3d.x(), max3d.y(), 0))
                              .addPoint(max3d)
                              .addPoint(Eigen::Vector3f(max3d.x(), min3d.y(), 0))
                              .color(simox::Color::gray()));
        }
    }

    void
    TebObstacleManager::addHumanObstacle(const human::Human& human, viz::Layer* visLayer)
    {
        auto proxemicZones = proxemics.createProxemicZones(human);

        int i = 0;
        for (const auto& proxemicZone : proxemicZones)
        {
            auto pose = conv::toRos(proxemicZone.pose);
            auto shape = conv::toRos(proxemicZone.shape);

            auto obst = boost::make_shared<teb_local_planner::extension::EllipseObstacle>(
                pose.position(), pose.theta(), shape.a, shape.b);

            const auto& penalty = proxemicZone.penalty;

            obst->setPenaltyModel(boost::make_shared<teb_local_planner::ExponentialPenaltyModel>(
                teb_local_planner::LinearPenaltyModel(penalty.minDistance, penalty.epsilon),
                penalty.exponent));

            obst->setWeight(proxemicZone.weight);
            obst->setHomotopicRelevance(proxemicZone.homotopicRelevance);

            if (human.linearVelocity != Eigen::Vector2f::Zero())
            {
                obst->setCentroidVelocity(conv::toRos2D(human.linearVelocity));
            }

            container.push_back(obst);

            // visualize proxemic zone if layer is available
            if (visLayer != nullptr)
            {
                const Eigen::Vector3f axisLength(proxemicZone.shape.a, proxemicZone.shape.b, 10.f-i);
                const core::Pose pose3d = conv::to3D(proxemicZone.pose);

                visLayer->add(viz::Ellipsoid("proxemicZone_" + std::to_string(visualizationIndex++))
                                  .pose(pose3d)
                                  .axisLengths(axisLength)
                                  .color(PROXEMIC_ZONE_COLOR[i % PROXEMIC_ZONE_COLOR.size()]));
            }
            i++;
        }
    }

} // namespace armarx::navigation::local_planning
